package org.vishia.odt.readOdt;

import java.util.LinkedList;
import java.util.List;

import org.vishia.util.StringFunctions;
import org.vishia.util.StringFunctions_C;

public class CodeIncludeInfo {

  final String sErrorMsg;
  
  final String sLineIncl;
  
  final String sFile;
  
  final String sTag;
  
  final int maxLineLen;
  
  final List<CodeIncludeInfo.MarkerLine> markerLines;
  
  /**
   * @param line Starting from the path after 'include:' till exclusive '.>',
   *        both usable for
   *         the CaptionCode paragraph text for odt2VML and also '&lt;:include:...' for VML2odt
   */
  public CodeIncludeInfo ( String sLineIncl) {
    int posPathStart = 8;                                  // before is 'include:' 8 character.
    int posPathEnd = sLineIncl.indexOf("::");              //     posPathEnd---------+                 +---posEnd
    int posEnd = sLineIncl.indexOf(".>", posPathEnd +2);   //  include:path/to/file.c::TAG::44::2:a,4:B.>
    if(posEnd <0) { posEnd = sLineIncl.indexOf('>', posPathEnd +2); } //fallback          |   | 2:a,4:B> 
    if(posEnd <0) { posEnd = sLineIncl.length(); }         // in odt                      |   |
    int posTagEnd = sLineIncl.indexOf("::", posPathEnd +2);//     posTagEnd---------------+   |
    if(posTagEnd <0) { posTagEnd = posEnd; }               //            posLenEnd------------+
    int posLenEnd = sLineIncl.indexOf("::", posTagEnd +2); //<:include:path/to/file.c::TAG::44.>
    if(posLenEnd <0) { posLenEnd = posEnd; }               //<:include:path/to/file.c::TAG.>
    this.sLineIncl = sLineIncl.substring(0, posEnd);
    if(posPathEnd <0 || posEnd <0) {
      this.sErrorMsg = String.format("\nERROR %s, '::TAG.>' missing", sLineIncl);
      this.sFile = this.sTag = null;
      this.maxLineLen = 0;
      this.markerLines = null;
      return;
    } else {
      this.sFile = sLineIncl.substring(posPathStart, posPathEnd).trim();  
      this.sTag = sLineIncl.substring(posPathEnd+2, posTagEnd).trim();
      if(posLenEnd > posTagEnd+2) {
        this.maxLineLen = StringFunctions_C.parseIntRadix(sLineIncl, posTagEnd+2, 99, 10, null);
      } else {
        this.maxLineLen = 44;
      }
      if(posEnd > posLenEnd) {
        int posM = posLenEnd +1;
        this.markerLines = new LinkedList<>();
        int[] len = new int[1];
        do {
          posM +=1;   //first ok, second: after ','
          CodeIncludeInfo.MarkerLine m = new CodeIncludeInfo.MarkerLine();
          m.line = StringFunctions_C.parseIntRadix(sLineIncl, posM, 99, 10, len);
          posM += len[0];
          if(sLineIncl.charAt(posM) == ':') {
            posM +=1;
            int pos2 = StringFunctions.indexOfAnyChar(sLineIncl, posM, posEnd, ",.x");
            if(pos2 <0) { pos2 = posEnd; }
            m.marker = sLineIncl.substring(posM, pos2);
            posM = pos2;
          }
          this.markerLines.add(m);
        } while(sLineIncl.charAt(posM) == ',');
      } else {
        this.markerLines = null;
      }
      this.sErrorMsg = null;
    }

  }
  
  
  static class MarkerLine { int line; String marker;}

}
