package org.vishia.odt.readOdt;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Map;
import java.util.TreeMap;

import org.vishia.msgDispatch.LogMessage;
import org.vishia.util.Arguments;
import org.vishia.util.Debugutil;
import org.vishia.util.FileFunctions;
import org.vishia.util.StringFunctions;

abstract class TranslateOdtCommon {

  
  
  protected static abstract class CommonArgs extends Arguments {
    
    
    /**The input file to read, either odt or VmL. */
    File dirIn, fIn;
    
    String sNameDoc, sExtIn;
    
    /**The input file to read, either odt or VmL. */
    File dirOut, fOut;
    
    String sExtOut;
    
    /**If set then create backup files for the current file.vml before writing the new one. 
     * 
     */
    File dirOutBack;
    

    /**If given as true by "-backall:dir", then rename all existing backup files with a number (as "file.99.vml") to preserve all versions.
     * Set to false by "-back:dir"
     */
    boolean bBackAll;
    
    /**This is either name and ext from the input markup file on read VmL or same for output file on write VmL.*/
    String sNameMarkup, sExtMarkup;
    

    
    /**Directory of the label file.*/
    public File dirLabel;
    
    /** "*" if * is written, read all files. */
    public String sNameLabel;
    

    /**Contains a relative (or possible absolute) or a URL path where foreign linked bookmark files are placed. 
     * It has the form "path/*.ext" where ".ext" may be ".html" to link to html files.
     */
    String sLinkForeignHtml, sExtForeignHtml, sLinkForeignPdf, sExtForeignPdf;
    

    File dirDbg;
    
    
    Argument[] argListCommon = {
        new Argument("-dirDbg", ":<dirWrTest>    directory for debug output", new SetArgument(){ 
          @Override public boolean setArgument(String val) throws FileNotFoundException { 
            CommonArgs.this.dirDbg = new File(val).getAbsoluteFile();  
            FileFunctions.mkDir(CommonArgs.this.dirDbg);
            return CommonArgs.this.dirDbg.exists();
          }})
      , new Argument("-labels", ":<path/*.label.txt>  file contains labels", new SetArgument(){ 
          @Override public boolean setArgument(String val) throws FileNotFoundException { 
            //int posWildcard = val.indexOf('*');
            int posName = val.lastIndexOf('\\')+1;
            int posName2 = val.lastIndexOf('/')+1;
            if(posName2 > posName) { posName = posName2; }
            //int posDot = val.indexOf('.', posName);
            //if(posWildcard <0) { 
            CommonArgs.this.sNameLabel = val.substring(posName); //} // maybe the "*"
            //CmdArgs.this.sExtLabel = val.substring(posDot);
            CommonArgs.this.dirLabel = new File(posName == 0 ? "." : val.substring(0, posName)).getAbsoluteFile();
            return CommonArgs.this.dirLabel.exists();
          }})
      , new Argument("-back", ":path/to  dir for backup vml files", new SetArgument(){ 
        @Override public boolean setArgument(String val) throws FileNotFoundException { 
          CommonArgs.this.dirOutBack = new File(val).getAbsoluteFile();  
          CommonArgs.this.bBackAll = false;
          return CommonArgs.this.dirOutBack.exists();
        }})
      , new Argument("-backall", ":path/to  dir for backup vml files, all are preserved with .99.vml", new SetArgument(){ 
        @Override public boolean setArgument(String val) throws FileNotFoundException { 
          CommonArgs.this.dirOutBack = new File(val).getAbsoluteFile();  
          CommonArgs.this.bBackAll = true;
          return CommonArgs.this.dirOutBack.exists();
        }})
      , new Argument("-alinkhtml", ":D:/path/*.html OR https://www.xyz.org/path/*.html path for foreign bookmark references", new SetArgument(){ 
        @Override public boolean setArgument(String val) throws FileNotFoundException { 
          int posWildcard = val.indexOf('*');
          CommonArgs.this.sLinkForeignHtml = val.substring(0, posWildcard);
          CommonArgs.this.sExtForeignHtml = val.substring(posWildcard +1);
          return true;
        }})
      , new Argument("-alinkpdf", ":D:/path/*.pdf OR https://www.xyz.org/path/*.pdf path for foreign bookmark references", new SetArgument(){ 
        @Override public boolean setArgument(String val) throws FileNotFoundException { 
          int posWildcard = val.indexOf('*');
          CommonArgs.this.sLinkForeignPdf = val.substring(0, posWildcard);
          CommonArgs.this.sExtForeignPdf = val.substring(posWildcard +1);
          return true;
        }})
      , new Argument("-rlinkhtml", ":../html/*.html OR https://www.xyz.org/path/*.html path for foreign bookmark references", new SetArgument(){ 
        @Override public boolean setArgument(String val) throws FileNotFoundException { 
          int posWildcard = val.indexOf('*');            // Note: the "../" is necessary from inside content.xml as relative path
          CommonArgs.this.sLinkForeignHtml = "../" + val.substring(0, posWildcard);
          CommonArgs.this.sExtForeignHtml = val.substring(posWildcard +1);
          return true;
        }})
      , new Argument("-rlinkpdf", ":../pdf/*.pdf OR https://www.xyz.org/path/*.pdf path for foreign bookmark references", new SetArgument(){ 
        @Override public boolean setArgument(String val) throws FileNotFoundException { 
          int posWildcard = val.indexOf('*');            // Note: the "../" is necessary from inside content.xml as relative path
          CommonArgs.this.sLinkForeignPdf = "../" + val.substring(0, posWildcard);
          CommonArgs.this.sExtForeignPdf = val.substring(posWildcard +1);
          return true;
        }})
  
    };

  }    

  
  final public LogMessage log;


  
  /**Index of Files, for which an index of all unique operation anchors are read already.
   * The inner Map is filled with {@link #createHtmlAnchorOperations(String)}
   */
  private Map<String, Map<String, String>> idxHtmlFilesAnchors = new TreeMap<>();
  private Object args;
  
  protected TranslateOdtCommon (LogMessage log) {
    this.log = log;
  }
  
  protected String searchForReplaceLinkOperationLabel( String sRef, int posFileEnd, File dirBase) {
    String retRef = sRef;
    String sHtml = sRef.substring(0, posFileEnd);
    if(sRef.contains("XmlSequWriter"))
      Debugutil.stop(); 
    Map<String, String> idxOperations = this.idxHtmlFilesAnchors.get(sHtml);
    try {
      if(idxOperations == null) {
        File fHtml = new File(dirBase, sHtml);
        if(fHtml.exists()) {
          idxOperations = createHtmlAnchorOperations(sHtml, dirBase);
          this.idxHtmlFilesAnchors.put(sHtml, idxOperations);
        } else {
          this.log.writeError("Warning: file for operation name check not found: '%s' ", fHtml.getAbsolutePath());
        }
      }
    } catch(IOException exc) {
      this.log.writeError("ERROR read html file '%s'", sHtml);
    }
    if(idxOperations !=null) {
      int posName = posFileEnd+1;
      int posNameEnd2 = StringFunctions.indexAfterIdentifier(sRef, posFileEnd+1, -1, null);
      //int posName = sRef.lastIndexOf('#', posNameEnd);
      //if(posName <0) { posName = sRef.lastIndexOf('.', posNameEnd); }
      String name = sRef.substring(posName, posNameEnd2); // Hint: posName is either -1 or position of '.' or '#'
      String sOperAnchor = idxOperations.get(name);
      if(sOperAnchor == null) {
        this.log.writeError("Warning: operation '%s' not found in: '%s' ", name, sHtml);
        retRef = sRef;
      } else if(sOperAnchor.endsWith("?")) {
        this.log.writeError("Warning: operation '%s' is not unique in: '%s' ", name, sHtml);
        retRef = sRef;
      } else {
        retRef = sHtml + '#' + sOperAnchor;
      }
    }
    return retRef;
  }
  
  
  private Map<String, String> createHtmlAnchorOperations (String sHtml, File dirBase) throws IOException{
    Map<String, String> idxOperations = new TreeMap<>();
    BufferedReader rHtml = new BufferedReader(new FileReader(new File(dirBase, sHtml)));
    String sLine;
    while((sLine = rHtml.readLine()) !=null) {
      int posEndName;
      if( sLine.startsWith("<a name=\"")                   // regard the structure of Javadoc result of Standard Java8
       && (posEndName = StringFunctions.indexAfterIdentifier(sLine, 9, -1, null)) >0
       && sLine.charAt(posEndName) == '-'                  // <a name="operationName- 
        ) {
        String operationName = sLine.substring(9, posEndName);
        int posAnchorEnd = sLine.indexOf('\"', posEndName);
        String sAnchor = sLine.substring(9, posAnchorEnd);
        String sExists = idxOperations.get(operationName);
        if(sExists !=null) {
          this.log.writeInfo("Info: operation twice: %s", sExists);
          idxOperations.put(operationName, sExists + "?"); // mark with "...?" for non unique operation
        } else {
          idxOperations.put(operationName, sAnchor);
        }
      }
    }
    rHtml.close();
    return idxOperations;
  }
  
  
  
  protected void writeBackupFile (ReadOdt.CommonArgs args) {
    if(args.fOut.exists() && args.dirOutBack !=null) {
      String sNameBack = args.sNameDoc;
      if(FileFunctions.getCanonicalPath(args.dirOutBack).equals(FileFunctions.getCanonicalPath(args.dirOut))) {
        sNameBack += ".back";
      }
      boolean bOkRename = true;
      if(args.bBackAll) {
        File fBack0 = new File(args.dirOutBack, sNameBack + args.sExtOut);
        if(fBack0.exists()) {
          int nr = 0;
          File fBack;
          do {
            String sFileBack = sNameBack + "." +  (++nr) + args.sExtOut;
            fBack = new File(args.dirOutBack, sFileBack);
          } while(fBack.exists());                     // search the first non exist file with nr.
          bOkRename = fBack0.renameTo(fBack);    // rename the last backup file with this number.
          if(bOkRename) { this.log.writef("\nsuccessfull backall: %s", fBack.getAbsolutePath()); }
          else { this.log.writef("\nERROR backall: %s", fBack.getAbsolutePath()); }
      } }
      //======>>>>                    =================== copy the vml back version.
      File fBack = new File(args.dirOutBack, sNameBack + args.sExtOut);
      try { 
        FileFunctions.copyFile(args.fOut, fBack);
        this.log.writef("\nsuccessfull back: %s", fBack.getAbsolutePath());
      } catch(IOException exc) {
        this.log.writef("\nERROR back: %s : %s", fBack.getAbsolutePath(), exc.getMessage());
      }
    }
    
  }


}
